# test_race_details.py

import unittest
from datetime import timedelta
from datetime import datetime
from horse_race_simulator.race_data.race_details import DelayedRace

class TestRace(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        print("Setting up resources for the entire TestRace class")
        cls.today_date = datetime.now()
        cls.today_date_str = datetime.now().strftime("%Y-%m-%d")
          
        cls.venue_names = ["Pony Speedway", "Canter Canyon", "Saddle Summit", "Dusty Lanes", "Gallop Galley", "Riders Run"]
        cls.venue_distances = [1000, 1200, 1600, 1800, 2200, 2400]
        cls.weather_conditions = ["Sunny", "Overcast", "Rainy", "Snowy"]

    @classmethod
    def tearDownClass(cls):
        print("Tearing down shared resources.")
        cls.today_date = None
        cls.today_date_str = None
        cls.venue_names = None
        cls.weather_conditions = None

    def setUp(self):
        print("Setting up testcase.")
        self.num_horses = 10
        self.race = DelayedRace(num_horses = self.num_horses)

    def tearDown(self):
        print("Tearing down testcase.")
        self.num_horses = None
        self.race = None
    
    def test_constructor(self): # To check default values
        print("Running test_constructor")
        # Check member variables during the initialization of DelayedRace
        self.assertIsNotNone(self.race.track)
        self.assertIn(self.race.venue, self.venue_names) # check if venue name is in list
        self.assertIn(self.race.distance, self.venue_distances) # check if venue distance is in list
        self.assertIn(self.race.weather, self.weather_conditions) # check if weather is in list

        current_date = (self.today_date + timedelta(days=self.race.delay_days)).strftime("%Y-%m-%d")
        self.assertEqual(self.race.date, current_date)
        self.assertGreater(self.race.prize, 4999)
        self.assertLessEqual(self.race.prize, 25000)
        self.assertEqual(self.race.num_horses, self.num_horses)
        self.assertIsNotNone(self.race.race_id)
        self.assertEqual(len(self.race.horses), self.num_horses)

        delay_days = 3
        self.assertEqual(self.race.delay_days, delay_days)

    def test_set_delayed_date(self):
        print("Running test_set_delayed_date")
        # Check values of delay_days and date before calling the set_delayed_date function
        delay_days = 3
        current_date = self.today_date + timedelta(days=delay_days)
        current_date_str = current_date.strftime("%Y-%m-%d")
        self.assertEqual(self.race.date, current_date_str)
        self.assertEqual(self.race.delay_days, delay_days)
        delay_days = 5
        self.race.delay_days = delay_days
        self.race.set_delayed_date()
        # Check values of delay_days and date after calling the set_delayed_date function
        new_date_str = (current_date + timedelta(days=(delay_days))).strftime("%Y-%m-%d")
        self.assertEqual(self.race.date, new_date_str)
        self.assertEqual(self.race.delay_days, delay_days)

    def test_set_date(self):
        print("Running test_set_date")
        # Check values before calling the function
        self.assertNotEqual(self.race.date, self.today_date_str, "It should not match today's date since set_delayed_date was called in the constructor.")
        self.race.set_date(self.today_date_str)
        # Check values after calling the function
        self.assertEqual(self.race.date, self.today_date_str)

    def test_get_race_info(self): # extra test, just making sure no exception raised when printing info
        print("Running test_get_race_info")
        self.race.get_race_info() 

if __name__ == "__main__":
    unittest.main(argv=[''], verbosity=2, exit=False)


# betting.py

from horse_race_simulator.simulation.race_simulator import RaceSimulation
from horse_race_simulator.simulation.race_results import RaceResults
from horse_race_simulator.race_data.race_details import DelayedRace


class User:
    """A class representing a User object.
    Methods:
        __init__(): Initializes 'User' instance.
        race_welcome(): Opening prompt for the game.
        run_game(): Consolidate the betting, race results and simulation into one method.
        show_balance(): Show user balance for the instance.
        take_bet(): Takes bet from user.
        distribute_earnings(): Distribute earnings after the race is completed.
    """
    def __init__(self, start_balance=1000):
        """
        Initializes instance of the 'User' class.

        Args:
            start_balance (int): Start balance of user, with default value = 1000.
        """
        self.balance = start_balance

    def race_welcome(self):
        """
        Opening prompt for the game.

        Args:
            self : Instance of the class.
        """
        print("Welcome to Horsle, a horse race simulator with turtles!")
        while True:
            start_check = input("Type 'start' to begin the race or 'exit' to quit: ").lower()
            if start_check == "exit":
                print("Thank you for racing! Goodbye!")
                break
            elif start_check == "start":
                self.run_game()
                if self.balance <= 0:
                    print("You have run out of money! Goodbye!")
                    break
                else:
                    print("Thank you for playing! Goodbye!")
                    break

    def run_game(self):
        """
        Consolidate the betting, race results and simulation into one method.

        Args:
            self : Instance of the class.
        """
        race = DelayedRace()
        race.get_race_info()
        track = race.track

        print("+" + "-" * 105 + "+")
        print("|                                         Horses in Today`s race:                                         |")
        print("+" + "-" * 105 + "+")

        for horse in race.horses:
            horse.get_horse_info()

        print("+" + "-" * 105 + "+")

        selected_horse_id = None
        bet = None
        while bet is None:
            bet, selected_horse_id = self.take_bet(race.horses)
            if bet is None:
                return

        # Run the race
        try:
            race_simulation = RaceSimulation(race, track)
            race_simulation.start_race()
            winning_horse_id = race_simulation.get_winning_horse_id()
            horse_times = race_simulation.get_times()
        except Exception as e:
            print(f"Error during race simulation: {e}")
            return

        # Display results and distribute earnings
        self.distribute_earnings(bet, winning_horse_id, selected_horse_id)
        self.show_balance()
        results = RaceResults(race, race.horses, horse_times)
        results.display_options()


    def show_balance(self):
        """
        Show user's starting balance for the instance.

        Args:
            self : Instance of the class.
        """
        print(f"Your balance is: ${self.balance:.2f}")

    def take_bet(self, horses):
        """
        Takes bet from user.

        Args:
            self : Instance of the class.
            horses (list) : List of Horse IDs for the current race.

        Returns:
            bet (int) : Bet value.
            horse_choice (int) : Horse ID of horse selected by the user.

        """

        while True:
            try:
                horse_choice = int(input(f"Choose a horse to place a bet on for today's race: "))

                valid_horses = [horse for horse in horses if horse.horse_id == horse_choice]
                if not valid_horses:
                    print(f"Horse {horse_choice} is not in today's race. Please select a valid horse.")
                    continue
                break
            except ValueError:
                print("Invalid input. Please enter a valid horse ID.")
                continue

        self.show_balance()

        while True:
            try:
                bet = float(input("How much would you like to bet? $"))

                if bet > self.balance:
                    print("Insufficient funds to place bet. Please enter a valid amount.")
                    continue
                elif bet <= 0:
                    print("Bet amount must be greater than zero. Please enter a valid bet.")
                    continue
                break
            except ValueError:
                print("Invalid input. Please enter a valid numeric bet amount.")
                continue

        self.balance -= bet

        print(f"You have placed a bet of ${bet:.2f} on horse {horse_choice}. The race will start shortly, good luck!\n")

        return bet, horse_choice

    def distribute_earnings(self, bet, winning_horse_id, selected_horse_id, odds=2.0):
        """
        Distribute earnings after the race is completed.

        Args:
            self : Instance of the class.
            bet (int) : Bet value.
            winning_horse_id (int) : Horse ID of winner of the race.
            selected_horse_id (int) : Horse ID of horse selected by the user.
            odds (int) : Odds of winning, default value set to 2.0.

        """
        if winning_horse_id == selected_horse_id:
            winnings = bet * odds
            self.balance += winnings
            print(f"\nCongratulations! Horse ID {selected_horse_id} won. You earned ${winnings:.2f}.")
        else:
            print(f"\nSorry, Horse ID {selected_horse_id} did not win. Your balance will be reduced by ${bet:.2f}.")



# race_results.py

import pandas as pd

class InvalidInputError(Exception):
    """Custom exception raised when input is not a valid input."""
    pass

class RaceResults:
    """A class representing the results of the race simulation
    Methods:
        __init__(): Intializes race result information
        get_horse_timing_data_frame(): retrieves the horse times during the race simulation (legs and overall time)
        get_horse_position(): finding the horse's position (used at various stages of the race)
        display_options(): gives users options to select the displaying of race results
        display_leaderboard(): displays a leaderboard of the race results
        generate_race_summary(): displays a detailed summary of the race
        get_horse_performance(): displays performance details for a specific horse
        get_horse_age(): retrieve horse age
        get_horse_type(): retrieve horse type (breed)
        get_horse_weight(): retrieve horse weight
        get_horse_jockey(): retrieve horse jockey (the person riding the horse)
    """
    
    hist = pd.read_csv('runs.csv')
    track_length = 50 # solely for display purposes

    def __init__(self, race, horses, horse_timings):
        """Initializes race results information"""
        self.race_id = race.race_id
        self.date = race.date
        self.horses = horses
        self.data = self.get_horse_timing_data_frame(horse_timings)

    def get_horse_timing_data_frame(self, horse_timings):
        """retrieves a data frame with the horse times across the race simulation
        Parameters:
           self: the race results object
           horse_timings: the dictionary of timing from the race simulation
        Returns:
           data frame: data frame of the horse ids witht their times
        """
        min_finish_time_overall = min(horse_timings.items(), key=lambda item: item[1]['Overall Time'])[1]['Overall Time'];
        min_finish_time_first_leg = min(horse_timings.items(), key=lambda item: item[1]['Leg 1 Time'])[1]['Leg 1 Time'];
        min_finish_time_second_leg = min(horse_timings.items(), key=lambda item: item[1]['Leg 2 Time'])[1]['Leg 2 Time'];
        min_finish_time_third_leg = min(horse_timings.items(), key=lambda item: item[1]['Leg 3 Time'])[1]['Leg 3 Time'];
        
        horse_ids = []
        step1_location = []
        step2_location = []
        step3_location = []
        step4_location = []
        final_position = []
        finish_times = []
        poisition = 1;

        for horse_id, timings in horse_timings.items():
            horse_ids.append(horse_id)
            step1_location.append(RaceResults.track_length * 0.25 * (min_finish_time_first_leg/timings['Leg 1 Time']))
            step2_location.append(RaceResults.track_length * 0.5 * (min_finish_time_second_leg/timings['Leg 2 Time']))
            step3_location.append(RaceResults.track_length * 0.75 * (min_finish_time_third_leg/timings['Leg 3 Time']))
            step4_location.append(RaceResults.track_length * (min_finish_time_overall/timings['Overall Time']))
            finish_times.append(timings['Overall Time'])
            final_position.append(self.get_horse_position(horse_timings, horse_id))
            poisition += 1

        horse_timing_data_set = {'horse_id': horse_ids, 'steps1': step1_location, 'steps2': step2_location, 'steps3': step3_location, 'steps4': step4_location,'result': final_position, 'finish_time': finish_times}
        return pd.DataFrame(horse_timing_data_set)

    def get_horse_position(self, horse_timings, horse_id):
        """retrieves a horses position
        Parameters:
           self: the race results object
           horse_timings: the set of horse times from the simulation 
           horse_id: a specific horse id (that was in the race)
        Returns:
           position: the horse's position, or -1 if the horse id is not found
        """

        sorted_horses = sorted(horse_timings.items(), key=lambda item: item[1]['Overall Time'])
    
        # Find the position of the specified horse
        for position, (current_horse_id, _) in enumerate(sorted_horses, start=1):
            if current_horse_id == horse_id:
                return position
    
        # If the horse_id is not found, return -1
        return -1
    
    def display_options(self):
        """retrieves a horses position
        Parameters:
           self: the race results object
        Returns:
           method calls: calls specific methods depending on user input
        """
        while True:
            try:
                results_type = input(
            """
        
If you would like to see an overview of race, please select one of the options below, otherwise select 'D':\n
A: Leaderboard
B: Overall summary
C: Compare horse performance
D: Exit\n
Enter your selection: """)
                if results_type == 'A':
                    self.display_leaderboard()
                elif results_type == 'B':
                    self.generate_race_summary()
                elif results_type == 'C':
                    self.get_horse_performance()
                elif results_type == 'D':
                    return None

                if results_type != 'A' and results_type != 'B' and results_type != 'C' and results_type != 'D':
                    raise InvalidInputError("\nInvalid input! please enter a valid input.")

            except InvalidInputError as e:
                # Handle the case where the invalid horse Id is given
                print(e)

    def display_leaderboard(self):
        """creates a visual display for the horses and their corresponding performance after the race.
        Parameters:
           self: the race results object
        Returns:
           print statements: The overall leaderboard from the race simulation and winner of the race
        """
           
        #Store in dataframe.
        sub_df = self.data[['result','horse_id','finish_time']].copy()
        sub_df = sub_df.sort_values(by='finish_time').copy().reset_index(drop=True)  
        position = sub_df['result']
        horse = sub_df['horse_id']
        finish_time = sub_df['finish_time']
        
        # Display the leaderboard
        print(f"\n🐢 Horse Race Leaderboard : {self.race_id}🐢")
        print("------------------------------------------------------------")
        print(f"{'Position':<10}{'Horse':<15}{'Time (s)':<10}")
        print("------------------------------------------------------------")
        for i in range(len(self.horses)):
            print(f"{position[i]:<10}{horse[i]:<15}{finish_time[i]:<10}")
    
        # Winner announcement
        print(f"\n🎉 Winner: {horse[0]} with a time of {finish_time[0]} seconds! 🎉")

    def generate_race_summary(self):
        """Display the race summary in detail includes attributes previously defined.
        Parameters:
           self: the race results object
        Returns:
           print statements: detailed race summary information
           """
        
        # Display the Performance results
        print(f"\n🐢 Race summary : {self.race_id} 🐢")
        print("------------------------------------------------------------")
        
        # Display all race attributes
        print(f" Date : {self.date}")
        print(f" Number of horses : {len(self.horses)}")
        print(f" Finish time of winner: {self.data['finish_time'].min()} seconds")
        print(f" Finish time of last horse: {self.data['finish_time'].max()} seconds")
         
        # Add visualizations
        
        horses = self.data['horse_id']
        positions = [0]* len(horses) # Starting positions
        
        print("\n🏁 Race Snapshot 🏁")
        print("------------------------------------------------------------")
        # Print race for each stage
        print('\nStage 1:\n')
        # Update positions 
        step1 = self.data['steps1'].astype(int)
        for i in range(len(self.data)):
            positions[i] = step1[i] 
        
        # Display the race track
        for i, horse in enumerate(horses):
            print(f"{horse}: " + "-" * positions[i] + "🐢" + "-" * (RaceResults.track_length - positions[i]))
        
        print('\nStage 2:\n')
        step2 = self.data['steps2'].astype(int)
        for i in range(len(self.data)):
            positions[i] = step2[i]  
        
        # Display the race track
        for i, horse in enumerate(horses):
            print(f"{horse}: " + "-" * positions[i] + "🐢" + "-" * (RaceResults.track_length - positions[i]))
        
        
        print('\nStage 3:\n')
        step3 = self.data['steps3'].astype(int)
        for i in range(len(self.data)):
            positions[i] = step3[i]  
        
        # Display the race track
        for i, horse in enumerate(horses):
            print(f"{horse}: " + "-" * positions[i] + "🐢" + "-" * (RaceResults.track_length - positions[i]))
        
        print('\nStage 4:\n')
        step4 = self.data['steps4'].astype(int)
        for i in range(len(self.data)):
            positions[i] = step4[i]  
        
        # Display the race track
        for i, horse in enumerate(horses):
            print(f"{horse}: " + "-" * positions[i] + "🐢" + "-" * (RaceResults.track_length - positions[i]))
        
    def get_horse_performance(self):
        """Retrieve performance summary for a specific horse, can be called for every horse in the race 
        Parameters:
           self: the race results object
        Returns:
           print statements: performance summary for a specific horse
        """

        sub_df = self.data[['result','horse_id']].copy()
        sub_df = sub_df.sort_values(by='result').copy().reset_index(drop=True)  
        position = sub_df['result']
        horse = sub_df['horse_id']
        
        print(f"🐢 Horse IDs 🐢")
        print("------------------------------------------------------------")
        print(f"{'Position':<10}{'Horse':<15}")
        print("------------------------------------------------------------")
        for i in range(len(self.horses)):
            print(f"{position[i]:<10}{horse[i]:<15}")

        horse_id_input = -1
        while True:
            try:
                take_input = input("\nEnter a horse id to compare performance with past results or enter to exit: ")
                if take_input == '':
                    return None

                horse_id_input = int(take_input)

                if self.find_horse_by_id(take_input) == None:
                    raise InvalidInputError("\nHorse id not in given list of horse ids.")

                break;
            except ValueError:
                # If the input cannot be converted to an integer, raise a custom exception
                print(f"\nInvalid input! '{take_input}' is not a valid horse id.")
            except InvalidInputError as e:
                # Handle the case where the invalid horse Id is given
                print(e)
        
        # Current race
        finish_time = self.data.loc[self.data['horse_id'] == horse_id_input,'finish_time'].iloc[0]
        rank = self.data.loc[self.data['horse_id'] == horse_id_input,'result'].iloc[0]
            
        # Historical data for comparison
        average_finish_time = RaceResults.hist[RaceResults.hist['horse_id'] == horse_id_input]['finish_time'].mean()
        win_count = RaceResults.hist[(RaceResults.hist['horse_id'] == horse_id_input) & (RaceResults.hist['won'] == 1)]['horse_id'].count()
        average_rank = round(RaceResults.hist[RaceResults.hist['horse_id'] == horse_id_input]['result'].mean())
        number_of_races = RaceResults.hist[RaceResults.hist['horse_id'] == horse_id_input]['race_id'].nunique()
        win_ratio = (win_count/number_of_races)*100
        
        # Display the Performance results
        print(f"🐢 Horse Performance : {horse_id_input} 🐢")
        print("------------------------------------------------------------")
        
        # Display all horse attributes
        print(f" {'Horse age:':<20} {self.get_horse_age(horse_id_input)}")
        print(f" {'Horse type:':<20} {self.get_horse_type(horse_id_input)}")
        print(f" Weight: {self.get_horse_weight(horse_id_input)}")
        print(f" Jockey: {self.get_horse_jockey(horse_id_input)}")
        print(f" Historical Win count: {win_count} of {number_of_races} races [Win ratio of {win_ratio:.2f}%] ")
    
        # Display table
        print("------------------------------------------------------------")
        print(f"{'Metric':<20}{'Current race':<20}{'Past performance':<20}")
        print("------------------------------------------------------------")
        print(f"{'Finish time (sec)':<20}{round(finish_time,5):<20}{round(average_finish_time,5):<20}")
        print(f"{'Rank':<20}{rank:<20}{average_rank:<20}")
        
    # supplementary methods
    def find_horse_by_id(self, horse_id):
        for horse in self.horses:
            if horse.horse_id == horse_id:
                return horse
        return None
    
    def get_horse_age(self, horse_id):
        for horse in self.horses:
            if horse.horse_id == horse_id:
                return horse.horse_age
        return None  # Return None if the horse_id is not found

    def get_horse_type(self, horse_id):
        for horse in self.horses:
            if horse.horse_id == horse_id:
                return horse.horse_type
        return None  # Return None if the horse_id is not found

    def get_horse_weight(self, horse_id):
        for horse in self.horses:
            if horse.horse_id == horse_id:
                return horse.actual_weight
        return None  # Return None if the horse_id is not found

    def get_horse_jockey(self, horse_id):
        for horse in self.horses:
            if horse.horse_id == horse_id:
                return horse.jockey_id
        return None  # Return None if the horse_id is not found



# horse_stats.py

from pandas import read_csv
from random import uniform

class Horse:
    """A class representing a Horse object.
    Methods:
        __init__(): Initializes 'Horse' instance.
        create_horse(): Retreives a subset of horses from the Kaggle data set.
        update_horse_stats(): Updates speed of a horse.
        get_horse_info(): Displays horse details.
    """

    def __init__(self, horse_id, horse_age, actual_weight, horse_type, horse_rating, jockey_id):
        """
        Initializes instance of the 'Horse' class.

        Args:
            horse_id (int): Horse ID.
            horse_age (int): Horse age.
            actual_weight (int): Actual weight of horse.
            horse_type (str): Horse type such as Colt, Mare, Gelding etc.
            horse_rating (int): Horse rating.
            jockey_id (int): Jockey ID.
        """

        self.horse_id = horse_id
        self.horse_age = horse_age
        self.actual_weight = actual_weight
        self.horse_type = horse_type
        self.horse_rating = horse_rating
        self.jockey_id = jockey_id
        self.speed = self.update_horse_stats()

    def create_horse(csv_filename):
        """
        Retreives a subset of horses from the given data set 'csv_filename'.

        Args:
            csv_filename: Kaggle data set containing horse data.
        """
        try:
            horse_df = read_csv(csv_filename)
        except FileNotFoundError:
            print(f"The file '{csv_filename}' was not found")
            return None
        try:
            horse_selection = horse_df.sample(n=1).iloc[0] # randomly taking one horse from the dataset
        except IndexError:
            print("Issue with dataset")
            return None
            
        horse_id = horse_selection['horse_id']
        horse_age = horse_selection['horse_age']
        actual_weight = horse_selection['actual_weight']
        horse_type = horse_selection['horse_type']
        horse_rating = horse_selection['horse_rating']
        jockey_id = horse_selection['jockey_id']

        horse_object = Horse(horse_id, horse_age, actual_weight, horse_type, horse_rating, jockey_id)

        return horse_object

    def update_horse_stats(self):
        """
        Updates speed of a horse based on factors; horse rating, horse age, actual weight of a horse and uses randomization to ensure that values vary with each simulation.

        Args:
            self: Instance of the class.
        """
        random_speed = uniform(30.0, 50.0)
        if self.horse_rating > 50:
            random_speed += 5
        if self.horse_age < 3:
            random_speed -= 2
        elif self.horse_age > 5:
            random_speed -= 5
        elif self.actual_weight < 125:
            random_speed += 5

        return round(random_speed, 2)

    def get_horse_info(self):
        """
        Prints out horse information.

        Args:
            self: Instance of the class.
        """
        print(
            f'| Horse ID: {self.horse_id:<4} | '
            f' Age: {self.horse_age} years | '
            f' Weight: {self.actual_weight} lbs | '
            f' Type: {self.horse_type:<8} | '
            f' Rating: {self.horse_rating:<3} | '
            f' Speed: {self.speed:.2f} km/h |'
        )
